from typing_extensions import TypeAlias

RangeLike = tuple[float, float]
Ranges = tuple[RangeLike, RangeLike, RangeLike]

# type alias
nm: TypeAlias = float
degree: TypeAlias = float
pixel: TypeAlias = int
subpixel: TypeAlias = float
